@@ -27,6 +27,15 @@
                  CHILD_STDIO_COUNT((buffer)) +      \
                  sizeof(HANDLE) * (fd)))
 
+#define CHILD_STDIO_COPY(buffer, dst, src) do {     \
+  DuplicateHandle(				    \
+    GetCurrentProcess(),		            \
+    CHILD_STDIO_HANDLE(buffer, src),		    \
+    GetCurrentProcess(),			    \
+    &CHILD_STDIO_HANDLE(buffer, dst),		    \
+    0, TRUE, DUPLICATE_SAME_ACCESS);		    \
+  } while (0)
+
 /* CRT file descriptor mode flags */
 #define FOPEN       0x01
 #define FEOFLAG     0x02
@@ -350,6 +359,16 @@ int processx__stdio_create(processx_handle_t *handle,
       if (err) { goto error; }
       CHILD_STDIO_CRT_FLAGS(buffer, i) = FOPEN | FDEV;
 
+    } else if (i == 2 && output[0] != '\0' && ! strcmp("2>&1", output)) {
+      /* This is stderr, sent to stdout */
+      /* We need to turn off buffering, otherwise the output on
+	 the two handles won't be correctly interleaved.
+	 We set FDEV on the pipes/files. This tricks windows
+	 into turning off the CRT buffering */
+      CHILD_STDIO_COPY(buffer, 2, 1);
+      CHILD_STDIO_CRT_FLAGS(buffer, 1) = FOPEN | FDEV;
+      CHILD_STDIO_CRT_FLAGS(buffer, 2) = FOPEN | FDEV;
+
     } else if (output[0] != '\0' && strcmp("|", output)) {
       /* output to file */
       if (i == 0) {
